//all credits to ArKano22 (of gamedev.net) for his implementation of
//Screen Space Ambient Occlusion:
//http://www.gamedev.net/community/forums/topic.asp?topic_id=556187&PageSize=25&WhichPage=1

// --------------------------------------------------------------------------------------------------
// PARAMETERS:
// --------------------------------------------------------------------------------------------------

//transforms
float4x4 tW: WORLD;        //the models world matrix
float4x4 tV: VIEW;         //view matrix as set via Renderer (EX9)
float4x4 tP: PROJECTION;
float4x4 tWVP: WORLDVIEWPROJECTION;
float4x4 tPI: PROJECTIONINVERSE; //:register(c%i);		// Full-fat 4x4 projection matrix that's been inverted, then transposed, then set row by row into these 4x constants

//texture
texture TexND <string uiname="Normal And Depth";>;
sampler SampleND = sampler_state    //sampler for doing the texture-lookup
{
    Texture   = (TexND);          //apply a texture to the sampler
    MipFilter = NONE;         //sampler states
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

texture TexRandomN <string uiname="Random Normals";>;
sampler SampRandomN = sampler_state    //sampler for doing the texture-lookup
{
    Texture   = (TexRandomN);          //apply a texture to the sampler
    MipFilter = NONE;         //sampler states
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

texture TexDiff <string uiname="Color Map";>;
sampler SampDiff = sampler_state    //sampler for doing the texture-lookup
{
    Texture   = (TexDiff);          //apply a texture to the sampler
    MipFilter = NONE;         //sampler states
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

float AO_Intensity <string uiname="AO Intensity";> = 1;
float GI_Intensity <string uiname="GI Intensity";> = 1;
float FarPlane <string uiname="Far Plane";> = 100;
float SampleRadius <string uiname="Sample Radius";> = 100;
float Width = 512;
float Height = 512;

float Jitter = 1;
float Scale = 1;

//the data structure: "vertexshader to pixelshader"
//used as output data with the VS function
//and as input data with the PS function
struct vs2ps
{
    float4 Pos  : POSITION;
    float2 TexCd : TEXCOORD0;
};

// --------------------------------------------------------------------------------------------------
// VERTEXSHADERS
// --------------------------------------------------------------------------------------------------
vs2ps VS(
    float4 PosO  : POSITION,
    float4 TexCd : TEXCOORD0)
{
    //declare output struct
    vs2ps Out;

    //transform position
    Out.Pos = mul(PosO, tWVP);
    
    //transform texturecoordinates
    Out.TexCd = TexCd;

    return Out;
}

// --------------------------------------------------------------------------------------------------
// PIXELSHADERS:
// --------------------------------------------------------------------------------------------------


float3 PosFromDepth(float2 UV)
{
  float4 sample = tex2D(SampleND, UV);
  float depth = (sample.b + sample.a / 255) * FarPlane;
  float4 pos = float4((UV.x - 0.5) * 2, (0.5 - UV.y) * 2, 1, 1);
  float4 ray = mul(pos, transpose(tPI));
  return ray.xyz * depth;
}

float3 ReadNormal(float2 UV)
{
  float4 sample = tex2D(SampleND, UV);
  float3 n;
  n.xy = sample.rg * 2 - 1;
  n.z = -sqrt(1 - dot(n.xy, n.xy));
  return n;
}

//Ambient Occlusion form factor:
float AOFF(float3 ddiff, float3 cnorm, float2 UV)
{
  float3 vv = normalize(ddiff);
  float rd = length(ddiff);
  return (1 - saturate(dot(ReadNormal(UV), -vv)))
         * saturate(dot(cnorm, vv))
         * (1 - 1 / sqrt(1 / (rd * rd * Scale) + 1));
}

//Global Illumination form factor:
float GIFF(float3 ddiff, float3 cnorm, float2 UV)
{
  float3 vv = normalize(ddiff);
  float rd = length(ddiff);
  return saturate(dot(ReadNormal(UV), -vv))
         * saturate(dot(cnorm, vv)) / (rd * rd + 1);
}

float4 PSAO(float2 UV: TEXCOORD0):COLOR
{
  //read current normal, position and color
  float3 currentNormal = ReadNormal(UV);
  float3 currentPosition = PosFromDepth(UV);
  // That intial value affects how "wooly" the randomiser is. Smaller numbers = grainer
  float3 random = 0.005 * SampleRadius * (tex2D(SampRandomN, UV * 5).xyz * 2 - 1);
  float3 currentDiffuse = tex2D(SampDiff, UV).rgb;

  float ao = 0;
  float3 gi = float3(0, 0, 0);
  float incx = 1 / Width * SampleRadius;
  float incy = 1 / Height * SampleRadius;
  float pw = incx;
  float ph = incy;

  float4 Sample = tex2D(SampleND, UV);
  float currentDepth = 1 / ((Sample.b + Sample.a / 255) * FarPlane);

  //3 rounds of 8 samples each
  float iterations = 3;
  for (float i=0; i<iterations; ++i)
  {
    float npw = (pw + random.x * Jitter) * currentDepth;
    float nph = (ph + random.y * Jitter) * currentDepth;

    float2 UV1 = UV + float2(npw, nph);
    float3 Dif1 = PosFromDepth(UV1) - currentPosition;
    float2 UV2 = UV + float2(npw, -nph);
    float3 Dif2 = PosFromDepth(UV2) - currentPosition;
    float2 UV3 = UV + float2(-npw, nph);
    float3 Dif3=  PosFromDepth(UV3) - currentPosition;
    float2 UV4 = UV + float2(-npw, -nph);
    float3 Dif4 = PosFromDepth(UV4) - currentPosition;
    float2 UV5 = UV + float2(0, nph);
    float3 Dif5 = PosFromDepth(UV5) - currentPosition;
    float2 UV6 = UV + float2(0, -nph);
    float3 Dif6 = PosFromDepth(UV6) - currentPosition;
    float2 UV7 = UV + float2(npw, 0);
    float3 Dif7 = PosFromDepth(UV7) - currentPosition;
    float2 UV8 = UV + float2(-npw, 0);
    float3 Dif8 = PosFromDepth(UV8) - currentPosition;

    ao += AOFF(Dif1, currentNormal, UV1);
    ao += AOFF(Dif2, currentNormal, UV2);
    ao += AOFF(Dif3, currentNormal, UV3);
    ao += AOFF(Dif4, currentNormal, UV4);
    ao += AOFF(Dif5, currentNormal, UV5);
    ao += AOFF(Dif6, currentNormal, UV6);
    ao += AOFF(Dif7, currentNormal, UV7);
    ao += AOFF(Dif8, currentNormal, UV8);

    pw += incx;
    ph += incy;
  }

  ao /= (iterations * 8);

  return float4(1 - saturate(ao.xxx * AO_Intensity), 1);
}

float4 PSAOGI(float2 UV: TEXCOORD0):COLOR
{
  //read current normal, position and color
  float3 currentNormal = ReadNormal(UV);
  float3 currentPosition = PosFromDepth(UV);
  // That intial value affects how "wooly" the randomiser is. Smaller numbers = grainer
  float3 random = 0.005F * SampleRadius * (tex2D(SampRandomN, UV * 5).xyz * 2 - 1);
  float3 currentColor = tex2D(SampDiff, UV).rgb;

  float ao = 0;
  float3 gi = float3(0, 0, 0);
  float incx = 1 / Width * SampleRadius;
  float incy = 1 / Height * SampleRadius;
  float pw = incx;
  float ph = incy;

  float4 Sample = tex2D(SampleND, UV);
  float currentDepth = 1 / ((Sample.b + Sample.a / 255) * FarPlane);

  //3 rounds of 8 samples each
  float iterations = 3;
  for (float i=0; i<iterations; ++i)
  {
    float npw = (pw + random.x) * currentDepth;
    float nph = (ph + random.y) * currentDepth;

    float2 UV1 = UV + float2(npw, nph);
    float3 Dif1 = PosFromDepth(UV1) - currentPosition;
    float2 UV2 = UV + float2(npw, -nph);
    float3 Dif2 = PosFromDepth(UV2) - currentPosition;
    float2 UV3 = UV + float2(-npw, nph);
    float3 Dif3=  PosFromDepth(UV3) - currentPosition;
    float2 UV4 = UV + float2(-npw, -nph);
    float3 Dif4 = PosFromDepth(UV4) - currentPosition;
    float2 UV5 = UV + float2(0, nph);
    float3 Dif5 = PosFromDepth(UV5) - currentPosition;
    float2 UV6 = UV + float2(0, -nph);
    float3 Dif6 = PosFromDepth(UV6) - currentPosition;
    float2 UV7 = UV + float2(npw, 0);
    float3 Dif7 = PosFromDepth(UV7) - currentPosition;
    float2 UV8 = UV + float2(-npw, 0);
    float3 Dif8 = PosFromDepth(UV8) - currentPosition;

    ao += AOFF(Dif1, currentNormal, UV1);
    ao += AOFF(Dif2, currentNormal, UV2);
    ao += AOFF(Dif3, currentNormal, UV3);
    ao += AOFF(Dif4, currentNormal, UV4);
    ao += AOFF(Dif5, currentNormal, UV5);
    ao += AOFF(Dif6, currentNormal, UV6);
    ao += AOFF(Dif7, currentNormal, UV7);
    ao += AOFF(Dif8, currentNormal, UV8);

    gi += GIFF(Dif1, currentNormal, UV1) * tex2D(SampDiff, UV1).rgb;
    gi += GIFF(Dif2, currentNormal, UV2) * tex2D(SampDiff, UV2).rgb;
    gi += GIFF(Dif3, currentNormal, UV3) * tex2D(SampDiff, UV3).rgb;
    gi += GIFF(Dif4, currentNormal, UV4) * tex2D(SampDiff, UV4).rgb;
    gi += GIFF(Dif5, currentNormal, UV5) * tex2D(SampDiff, UV5).rgb;
    gi += GIFF(Dif6, currentNormal, UV6) * tex2D(SampDiff, UV6).rgb;
    gi += GIFF(Dif7, currentNormal, UV7) * tex2D(SampDiff, UV7).rgb;
    gi += GIFF(Dif8, currentNormal, UV8) * tex2D(SampDiff, UV8).rgb;

    pw += incx;
    ph += incy;
  }

  ao /= (iterations * 8);
  gi /= (iterations * 8);

  return float4(saturate(currentColor + gi * GI_Intensity) - saturate(ao.xxx * AO_Intensity), 1);
}


// --------------------------------------------------------------------------------------------------
// TECHNIQUES:
// --------------------------------------------------------------------------------------------------

technique TSSAO
{
    pass P0
    {
        //Wrap0 = U;  // useful when mesh is round like a sphere
        VertexShader = compile vs_3_0 VS();
        PixelShader  = compile ps_3_0 PSAO();
    }
}

technique TSSAOGI
{
    pass P0
    {
        //Wrap0 = U;  // useful when mesh is round like a sphere
        VertexShader = compile vs_3_0 VS();
        PixelShader  = compile ps_3_0 PSAOGI();
    }
}
